/**
 * anymap-ts: Interactive maps with anywidget and TypeScript.
 *
 * Main entry point for the package.
 */

export * from './types';
export * from './core';

// MapLibre is the default export
export { default } from './maplibre';
